## package ardi.driver.historianipc
#
#  Provides Historian Driver HTTP IPC
#
#  Classes in this namespace provide inter-process-communication between historian drivers
#  and ARDI through a TCP socket mirroring HTTP.
try:
	import SocketServer
except:
	import socketserver as SocketServer
	
try:
    from urllib.parse import parse_qsl
except:
    from urlparse import parse_qsl

from threading import Thread

## Function used to generate a query in a thread.
def QueryFunction(socket,inst,cmd):
    print("Query Thread Started")
    socket.Respond(inst,cmd)
    print("Query Thread Stopped")

## Historian IPC Message Handler
#
#    This class responds to commands that are issued from server applications such as ARDI.
#
#    Some commands operate on the driver as a whole, while others operate on a single core
class IPCHandler(SocketServer.StreamRequestHandler):

    ## Handles Incoming Commands
    #
    #    This function reads from the incoming stream and response appropriately."""
    def handle(self):
        ln = self.rfile.readline()
        try:
                self.data = ln.decode('utf-8').strip()
        except:
                self.data = ln.strip()
                
        cmd = self.data.split('|')
        context = 1
        if len(cmd)>2:
            context = cmd[2]
            
        #print("Incoming Command: " + str(cmd))
        try:            
            if cmd[0] == 'spinup':
                if len(cmd) > 1:
                    self.server.core.Spinup(cmd[1],context)
                    return
            if cmd[0] == 'reload':
                if len(cmd) > 1:
                    self.server.core.Reload(cmd[1],context)
                    return
                else:
                    self.server.core.ReloadAll()
            if cmd[0] == 'spindown':
                self.server.core.Restart()
                return
            if cmd[0] == 'restart':
                self.server.core.Restart()
                return
            if cmd[0] == 'query':
                #Rebuild query...
                strx = ""
                for x in range(2,len(cmd)):
                    if x > 2:
                        strx = str + "|"
                    strx = cmd[x]

                ast = 0
                ctx = 1
                bits = cmd[1].split("_")
                if len(bits) > 1:
                        ast = bits[0]
                        ctx = int(bits[1])
                else:
                        ast = cmd[1]
                    
                self.ThreadRespond(ast,strx,context=ctx)
                return
            if cmd[0] == 'status':                
                stat = self.server.core.ShowStatus()
                self.wfile.write(stat.encode())
            if cmd[0] == 'log':
                cr = 0
                if len(cmd) > 1:                        
                        cr = int(cmd[1])                
                stat = self.server.core.ShowLog(cr,context,'basic')
                self.wfile.write(stat.encode())
            if cmd[0] == 'dlog':
                cr = 0
                if len(cmd) > 1:
                        cr = int(cmd[1])
                stat = self.server.core.ShowLog(cr,context,'detail')
                self.wfile.write(stat.encode())
            if cmd[0] == 'issues':
                cr = 0
                if len(cmd) > 1:
                        cr = int(cmd[1])
                stat = self.server.core.ShowIssues(cr)
                self.wfile.write(stat.encode())
        except:
            pass

        if self.data[0:14] == "GET /query?id=":
            print("Got Incoming")
            #Get the ID number...
            pos = self.data.find("&")
            if pos > 0:
                id = self.data[14:pos]
                ctx = 1
                bits = id.split('_')
                if len(bits) > 1:
                        try:
                                ctx = int(bits[1])
                        except:
                                pass
                self.ThreadRespond(id,self.data[pos+1:len(self.data)-9],context=ctx)

    ## Threaded response handler
    def ThreadRespond(self,instance,cmd,context=1):
        self.Respond(instance,cmd,context=context)
        #reactor.callInThread(QueryFunction,self,instance,cmd)

    ## Sends a query response to the client
    #
    #    This function issues a query via the driver and returns the results.
    #   
    #    \param instance The ARDI source asset ID
    #    \param cmd The query command string"""
    def Respond(self,instance,cmd,context=1):
        
        #Split Query String
        #print("Querying...")

        vals = parse_qsl(cmd)
        
        points = ""
        start = ""
        end = ""
        func = "raw"
        grain = 0
        first = 1        

        parameters = {}
        
        params = cmd.split("&")
        for x in vals:  
            #r = x.split("=")
            if x[0] == "start":
                parameters["start"] = x[1]                
            if x[0] == "end":
                parameters["end"] = x[1]
            if x[0] == "search":
                parameters["search"] = x[1]
            if x[0] == "function":
                parameters["function"] = x[1]
            if x[0] == "offset":
                parameters["offset"] = x[1]
            if x[0] == "count":
                parameters["count"] = x[1]
            if x[0] == "context":
                context = r[1]

        print(str(params))

        #print "Requesting History..."
        result = self.server.core.Query(instance,context,parameters)
        #print "Sending Response: " + result
        self.wfile.write("HTTP/1.1 200 OK\n".encode())
        self.wfile.write("Server: ARDI Consolidator\n".encode())
        self.wfile.write("Content-Type: text/plain\n".encode())
        self.wfile.write("Connection: Closed\n\n".encode())
        self.wfile.write(result)

## A Twisted mixin, creating a threaded TCP server.
class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    pass

## Historical IPC (Socket) server to control a driver
#
#    This class listens for socket connections and responds to commands from controlling applications (such as ARDI)
#
#    Some commands operate on the driver as a whole, while others operate on a single core
class DriverIPC:
    def __init__(self):
        """Intialises the IPC Connection"""

        ##The TCPServer instance that represents this service
        self.service = None

        ##The port that this IPC service should run on
        self.port = 9001

        ##A reference to the drivercore that spawned this IPC service
        self.core = None        

    def start(self):
        """Start the IPC Service

        This command starts the TCPServer that listens for incoming commands"""
        SocketServer.TCPServer.allow_reuse_address = 1
        #self.service = SocketServer.TCPServer(("",self.port),IPCHandler)
        self.service = ThreadedTCPServer(("",self.port),IPCHandler)
        self.service.core = self.core
        self.service.serve_forever()    

#IPC = DriverIPC()
#IPC.start()
